/*
 *  
 *  Copyright (c) 2004 Steve Slaven, All Rights Reserved.
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA 02111-1307 USA
 *  
*/
/* This is the app singleton class */

package net.hoopajoo.cheeselab3;

import javax.imageio.*;
import javax.imageio.stream.*;
import javax.swing.*;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;

import net.hoopajoo.swingx.*;

import java.awt.*;
import java.io.*;
import java.net.*;
import java.awt.event.*;
import java.awt.image.*;
import java.util.*;
import java.lang.reflect.*;

public class CheeseMap {
    static CheeseProgress progress = null;
    static String cachedir = "cache/_null";
    static int edit_roomid = 0;
    static JFrame toplevel = null;
    static Socket sock = null;
    static BufferedInputStream in = null;
    static PrintStream out = null;
    static CheeseCanvas canvas = null;
    static CheeseMapData map = null;
    static CheeseMapRoomEditor room_editor = null;
    static CheeseMapNavigator navi = null;
    static CheeseMapSearch find = null;
    static CheeseMapFollowDialog select_follow = null;
    static JToolBar toolbar = null;
    static CheeseMapToolbox toolbox = null;
    static volatile boolean online = false;


    static SelectedAction connect_action;
    static SelectedAction find_action;
    static SelectedAction navi_action;
    static SelectedAction ontop_action;

    /* This stuff should be canvas specific. Moved to Canvas class */
    //static String follow_userid = "";
    //static boolean follow_player = true;
    static boolean did_refresh = false;

    static volatile boolean pinger = false;

    static Properties props = null;

    static NetworkBonanza bozo = null;
    static int last_acceptid, last_sendid;

    // Listener vector for server events
    static Vector listeners = new Vector();

    private static void createAndShowGUI() {
        // Some initialization code in canvas needs these
        load_properties();
        apply_properties();

        //JFrame.setDefaultLookAndFeelDecorated( true );
        JFrame.setDefaultLookAndFeelDecorated( false );
        toplevel = new JFrame( "CheeseMap 3" );
        toplevel.setDefaultCloseOperation( JFrame.DO_NOTHING_ON_CLOSE );
        toplevel.addWindowListener( new WindowAdapter() {
            public void windowClosing( WindowEvent e ) {
                super.windowClosing( e );
                CheeseMap.exit();
            }
         } );

        Container content = toplevel.getContentPane();
        content.setLayout(new AKDockLayout());

        /* Map, we absolutely want to instantiate this before any canvases because
         * then events will be flowing in the correct order and cache will be up to
         * date when it's the canvases turn to get the event.
         */ 
        map = new CheeseMapData();

        // Canvas
        canvas = new CheeseCanvas();
        content.add( canvas, AKDockLayout.CENTER );

        // Restore rendering settings
        canvas.set_render_grid( prop_get_boolean( "render_grid", true ) );
        canvas.set_render_future( prop_get_boolean( "render_futures" ) );
        canvas.set_render_notes( prop_get_boolean( "render_notes" ) );

        // Create actions
        connect_action = new SelectedAction() {
            public void actionPerformed(ActionEvent e) {
                String cmd = e.getActionCommand();
                connect_action.setSelected( !isSelected() );
                if( cmd.equals( "connect" ) ) {
                    CheeseLab.dmesg( 3, "Connecting" );
                    connect();
                }else if( cmd.equals( "disconnect" ) ) {
                    CheeseLab.dmesg( 3, "Disconnecting" );
                    disconnect();
                }
            }
        };

        find_action = new SelectedAction() {
            public void actionPerformed(ActionEvent e) {
                CheeseMap.find.setVisible( true );
            }
        };
        find_action.putValue( SelectedAction.SHORT_DESCRIPTION, "Find" );
        find_action.putValue( SelectedAction.ACCELERATOR_KEY, KeyStroke
                .getKeyStroke( KeyEvent.VK_F, InputEvent.CTRL_MASK ) );

        navi_action = new SelectedAction() {
            public void actionPerformed(ActionEvent e) {
                setSelected( !isSelected() );
                CheeseMap.navi.setVisible( isSelected() );
            }
        };
        navi_action.putValue( SelectedAction.SHORT_DESCRIPTION, "Navigator" );
        navi_action.putValue( SelectedAction.ACCELERATOR_KEY, KeyStroke
                .getKeyStroke( 'n' ) );

        /* For win32 users since they don't have this in their WM */
        ontop_action = new SelectedAction() {
            public void actionPerformed(ActionEvent e) {
                setSelected( !isSelected() );
                // FIXME: When system takes our always on top status away
                // visuals will not be updated
                setAlwaysOnTop( isSelected() );
            }
        };
        ontop_action.putValue( SelectedAction.SHORT_DESCRIPTION, "Always On Top" );

        // Create toolbox
        toolbox = new CheeseMapToolbox();

        // Create toolbar
        toolbar = new JToolBar();

        // Setup toolbar
        JButton bb;
        JSelectedButton sb;

        // "Generic" commands (connect, follow etc.)
        sb = new JSelectedButton( connect_action );
        sb.setIcon( new ImageIcon( "ui/ui_disconnect.png" ) );
        sb.setSelectedIcon( new ImageIcon( "ui/ui_connect.png" ) );
        toolbar.add( sb );
        set_connect_status( false );

        bb = new JButton( find_action );
        bb.setIcon( new ImageIcon( "ui/ui_find.png" ) );
        toolbar.add( bb );

        sb = new JSelectedButton( canvas.follow_action );
        sb.setIcon( new ImageIcon( "ui/ui_follow.png" ) );
        toolbar.add( sb );

        // Zooming
        toolbar.addSeparator();
        bb = new JButton( canvas.zoomin_action );
        bb.setIcon( new ImageIcon( "ui/ui_zoomin.png" ) );
        toolbar.add( bb );
        bb = new JButton( canvas.zoomout_action );
        bb.setIcon( new ImageIcon( "ui/ui_zoomout.png" ) );
        toolbar.add( bb );

        // Navigation
        toolbar.addSeparator();
        bb = new JButton( canvas.levelup_action );
        bb.setIcon( new ImageIcon ( "ui/ui_u.png" ) );
        toolbar.add ( bb );

        bb = new JButton( canvas.leveldown_action );
        bb.setIcon( new ImageIcon ( "ui/ui_d.png" ) );
        toolbar.add ( bb );

        // Add toolbars to main window
        toggleToolbar(toolbar, prop_get_boolean( "show_toolbar", true ),
                      prop_get_string( "toolbar_location", AKDockLayout.NORTH ) );
        toggleToolbar(toolbox, prop_get_boolean( "show_toolbox", true ),
                      prop_get_string( "toolbox_location", AKDockLayout.NORTH ) );

        // Setup menus
        JMenuBar bar = new JMenuBar();

        // File menu
        JMenu menu = new JMenu( "File" );
        menu.setMnemonic( KeyEvent.VK_F );
        JMenuItem item;

        menu.add( new JSelectedMenuItem( connect_action ) );

        item = new JMenuItem( "Preferences", 'P' );
        item.addActionListener( new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                CheeseMap.show_preferences();
            }
        } );
        menu.add( item );
        item = new JMenuItem( "Change Follow", 'F' );
        item.addActionListener( new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                CheeseMap.select_follow.setVisible( true );
            }
        } );
        menu.add( item );
        item = new JMenuItem( "Change Player Icon", 'I' );
        item.addActionListener( new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                CheeseMap.change_avatar();
            }
        } );
        menu.add( item );
        item = new JMenuItem( "Sync Cache", 'H' );
        item.addActionListener( new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                CheeseMap.do_refresh();
            }
        } );
        menu.add( item );
        menu.addSeparator();
        item = new JMenuItem( "Exit", 'X' );
        item.addActionListener( new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                CheeseMap.exit();
            }
        } );
        menu.add( item );
        bar.add( menu );

        // Editor menu
        menu = new JMenu( "Edit" );
        menu.setMnemonic( 'E' );
        menu.add( new JSelectedMenuItem( CheeseMap.find_action ) );
        menu.add( toolbox.createMenu( "Tools" ) );
        menu.addSeparator();
        item = new JMenuItem( "Merge All", 'A' );
        item.addActionListener( new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                CheeseMap.canvas.merge_all();
            }
        } );
        menu.add( item );
        bar.add( menu );

        // Editor menu
        menu = new JMenu( "View" );
        menu.setMnemonic( 'V' );
        JMenu submenu = new JMenu( "Toolbars" );
        item = new JCheckBoxMenuItem( "Navigation Toolbar", toolbar.isVisible() );
        item.addActionListener( new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                toggleToolbar( toolbar, ((JCheckBoxMenuItem)e.getSource())
                        .isSelected() );
            }
        } );
        submenu.add( item );
        item = new JCheckBoxMenuItem( "Editor Toolbar", toolbox.isVisible() );
        item.addActionListener( new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                toggleToolbar( toolbox, ((JCheckBoxMenuItem)e.getSource())
                        .isSelected() );
            }
        } );
        submenu.add( item );
        menu.add( submenu );

        menu.add( new JSelectedCheckMenuItem( navi_action ) );
        menu.add( new JSelectedCheckMenuItem( canvas.follow_action ) );
        menu.add( new JSelectedMenuItem( canvas.levelup_action ) );
        menu.add( new JSelectedMenuItem( canvas.leveldown_action ) );
        menu.add( new JSelectedMenuItem( canvas.zoomin_action ) );
        menu.add( new JSelectedMenuItem( canvas.zoomout_action ) );

        item = new JCheckBoxMenuItem( "Render Notes", canvas.get_render_notes() );
        item.addActionListener( new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                CheeseMap.canvas.set_render_notes( ((JCheckBoxMenuItem)e
                        .getSource()).isSelected() );
                CheeseMap.canvas.repaint();
            }
        } );
        menu.add( item );

        item = new JCheckBoxMenuItem( "Render Futures", canvas
                .get_render_future() );
        item.addActionListener( new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                CheeseMap.canvas.set_render_future( ((JCheckBoxMenuItem)e
                        .getSource()).isSelected() );
                CheeseMap.canvas.repaint();
            }
        } );
        menu.add( item );

        item = new JCheckBoxMenuItem( "Render Grid", canvas.get_render_grid() );
        item.addActionListener( new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                CheeseMap.canvas.set_render_grid( ((JCheckBoxMenuItem)e
                        .getSource()).isSelected() );
                CheeseMap.canvas.repaint();
            }
        } );
        menu.add( item );

        bar.add( menu );

        // Window menu
        menu = new JMenu( "Window" );
        menu.setMnemonic( 'W' );

        menu.add( new JSelectedCheckMenuItem( ontop_action ) );

        item = new JMenuItem( "New Friend View" );
        item.addActionListener( new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                (new CheeseMapFriendView()).setVisible( true );
            };
        } );
        menu.add( item );
        menu.addSeparator();

        bar.add( menu );

        // Help menu
        menu = new JMenu( "Help" );
        menu.setMnemonic( 'H' );

        item = new JMenuItem( "Go to CheeseLab" );
        item.addActionListener( new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                // showURL("http://www.hoopajoo.net/mud/cheeselab.html");
            }
        } );
        // menu.add( item );

        item = new JMenuItem( "System Information" );
        item.addActionListener( new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                Runtime r = Runtime.getRuntime();
                String sysinfo = "Java version: "
                        + System.getProperty( "java.version" ) + "\n"
                        + "Memory used: " + (r.totalMemory() - r.freeMemory())
                        / 1024 / 1024 + " of " + r.maxMemory() / 1024 / 1024
                        + " MB\n\n";
                sysinfo += map.get_sysinfo();
                sysinfo += CheeseRendererCached.get_sysinfo();
                sysinfo += CheeseRendererThreaded.get_sysinfo();
                JOptionPane.showMessageDialog( toplevel, sysinfo,
                        "CheeseMap System Information",
                        JOptionPane.INFORMATION_MESSAGE );
            }
        } );

        menu.add( item );

        menu.addSeparator();

        item = new JMenuItem( "About" );
        item.addActionListener( new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                /*
                 * Default text to display if the about file is missing
                 */
                String text = "<pre>CheeseLab MUME Mapping tools \n\n"
                        + "Version: "
                        + CheeseLab.version()
                        + "\n\n"
                        + "(c) Copyright Steven Slaven and contributors 2004 - 2005\n\n"
                        + "Visit: http://www.hoopajoo.net/mud/cheeselab.html\n\n</pre>";
                try {
                    StringBuffer contents = new StringBuffer();
                    BufferedReader in = new BufferedReader( new FileReader(
                            "ui/about.html" ) );
                    String line = null;
                    while ((line = in.readLine()) != null) {
                        contents.append( line );
                        contents
                                .append( System.getProperty( "line.separator" ) );
                    }
                    text = contents.toString();
                    File cwd = new File( "." );
                    String dir = cwd.getCanonicalPath() + "/ui/";
                    CheeseLab.dmesg( 3, "Got canonical path: " + dir );

                    text = text.replaceAll( "@@VERSION@@", CheeseLab.version() );

                    /*
                     * We do this through a butthole to maintain compatibility
                     * with java 1.4. NO! We can't use replaceAll because it
                     * will crash when it gets some windows path names with \-s
                     * in them
                     */
                    String w = "@@BASE@@";
                    int t = text.indexOf( w );
                    text = text.substring( 0, t ) + "file:///" + dir
                            + text.substring( t + w.length(), text.length() );

                    // TODO: Remove the madness above and uncomment this neat
                    // line
                    // text = text.replace("@@BASE@@", "file:///" + dir );
                } catch (IOException ex) { }
                JEditorPane jp = new JEditorPane( "text/html", text );
                jp.setEditable( false );
                jp.setBackground( toplevel.getBackground() );
                jp.addHyperlinkListener( new HyperlinkListener() {
                    public void hyperlinkUpdate(HyperlinkEvent e) {
                        if( e.getEventType() == HyperlinkEvent.EventType.ACTIVATED ) {
                            // showURL(e.getURL());
                        }
                    }
                } );
                JOptionPane.showMessageDialog( toplevel, new JScrollPane( jp ),
                        "About CheeseMap", JOptionPane.PLAIN_MESSAGE );
            };
        } );

        menu.add( item );

        bar.add( menu );

        toplevel.setJMenuBar( bar );

        toplevel.pack();

        Toolkit tk = Toolkit.getDefaultToolkit();
        Dimension dm = tk.getScreenSize();

        /*
         * Restore toplevel window position and size from saved properties
         * Revert to 0,0, screenwidth, screenheight / 3 if not found or out of
         * visible area
         */
        Rectangle toplev_pos = new Rectangle( prop_get_int( "mainwin_x" ),
                prop_get_int( "mainwin_y" ), prop_get_int( "mainwin_w" ),
                prop_get_int( "mainwin_h" ) );

        if( 	toplev_pos.x < -toplev_pos.width + 20 ||
                toplev_pos.y < -toplev_pos.height + 20 ||
                toplev_pos.width <= 100 || toplev_pos.height <= 100 ||
                toplev_pos.x > dm.getWidth() - 20 ||
                toplev_pos.y > dm.getHeight() - 20 ||
                toplev_pos.x + toplev_pos.width > dm.getWidth() + toplev_pos.width - 20 ||
                toplev_pos.y + toplev_pos.height > dm.getHeight() + toplev_pos.height - 20 ) {
            toplevel.setSize( (int)dm.getWidth(), (int)dm.getHeight() / 3 );
        }else{
            toplevel.setBounds( toplev_pos );
        }

        // Check and restore tile size
        int tilesize = prop_get_int( "tile_size" );
        if( tilesize >= 1 && tilesize <= dm.getWidth() &&
                tilesize <= dm.getHeight() ) {
            set_tile_size( tilesize );
        }

        toplevel.setVisible( true );

        ontop_action.setSelected( prop_get_boolean( "mainwin_ontop" ) );
        setAlwaysOnTop( ontop_action.isSelected() );

        set_connect_status ( false );
        canvas.set_follow_player( true );

        // load the room editor
        room_editor = new CheeseMapRoomEditor( toplevel );
        find = new CheeseMapSearch( toplevel );
        navi = new CheeseMapNavigator( toplevel );
        progress = new CheeseProgress( toplevel );
        select_follow = new CheeseMapFollowDialog( toplevel );

        // restore navigator state
        // FIXME: There must be a cleaner way to do this
        navi_action.setSelected( prop_get_boolean( "show_navigator" ) );
        navi.setVisible( navi_action.isSelected() );
    }

    /** Ok, this is another through butthole solution for maintining
     * compatibility with Java 1.4. Basically we use the reflection interface
     * to look for our beloved setAlwaysOnTop method in JFrame class.
     * This way the code will compile with Java 1.4 and will actually do something
     * useful when run with Java 1.5 classes.
     * Cool, huh? Err ...... Grrr ......
     * 
     * TODO: Tidy this up when Steven decides to ditch java 1.4 support,
     *       I know it will happen eventually :)
     * @param b
     */
    private static void setAlwaysOnTop(boolean b) {
        Class c = toplevel.getClass();
        try {
            Method m = c.getMethod("setAlwaysOnTop", new Class [] { boolean.class } );
            m.invoke( toplevel, new Object [] { new Boolean( b ) } );
        } catch (Exception e) {
            // If we get here we are f....ed, disable the control
            ontop_action.setSelected( false );
            ontop_action.setEnabled( false );
        }
    }

    public static void set_edit_room( int roomid ) {
        set_edit_room( roomid, true );
    }

    public static void set_edit_room( CheeseRoomData r ) {
        set_edit_room( r, true );
    }

    public static void set_edit_room( int roomid, boolean open_editor ) {
        CheeseLab.dmesg( 2, "setting edit room# to " + roomid );
        // Make sure it gets in the cache
        map.get_room( roomid );

        // TODO: make sure we get the latest?
        CheeseMap.clearping( "set_edit_room" );
        send_tx( 1, "ping", "set_edit_room" );
        while( ! CheeseMap.checkping() );

        set_edit_room( map.get_room( roomid ), open_editor );
    }

    public static void set_edit_room( CheeseRoomData r, boolean open_editor ) {
        edit_roomid = r.roomid;
        canvas.center_on( r.x, r.y, r.z );
        room_editor.set_room( r );
        canvas.set_follow_player( false );
        if( open_editor )
            room_editor.setVisible( true );
    }

    public static void save_edit_room() {
        CheeseRoomData r = room_editor.get_room();
        save_room( r );
    }

    /* For now we execute all room data manipulation functions in a synchronized fashion.
     * I'm not sure if it's absolutely necessary, but better safe than sorry.
     */

    synchronized public static void save_room( CheeseRoomData r ) {
        if( r == null ) // Nothing to do
            return;

        CheeseRoomData oldroom = map.get_room( r.roomid, 1, false );

        send_tx( 1, "save", CheeseDecoder.encode_object( r ) );

        // Fire events
        for( int i = 0; i < listeners.size(); i++ ) {
            ((CheeseMapListener)listeners.elementAt( i )).roomChanged( r, oldroom );
        }
    }

    synchronized public static void delete_room( int roomid ) {
        CheeseRoomData room = map.get_room( roomid, 1, false );

        if( room == null ) // No such room (shouldn't happen)
            return;

        send_tx( 1, "delete", roomid + "" );

        // Fire events
        for( int i = 0; i < listeners.size(); i++ ) {
            ((CheeseMapListener)listeners.elementAt( i )).roomDeleted( room );
        }
    }

    synchronized public static void move_rooms( Integer [] roomids, int x, int y, int z ) {
        CheeseLab.dmesg( 3, "Gonna move rooms: " + x + "," + y + "," + z );

        // Loop over selected rooms and send save commands
        String tx = x + "," + y + "," + z;
        for( int i = 0; i < roomids.length; i++ ) {
            int roomid = roomids[ i ].intValue();
            CheeseLab.dmesg( 4, "  > " + roomid );
            tx += "," + roomid;
        }

        CheeseMap.send_tx( 1, "move", tx );

        // FIXME: Implement events

        // TODO: make sure if we are EDITING one of these that we update the room editor!
        // it might be easeier if the room tx handler checked to see if the room is in the
        // editor and just passed it on magically
    }

    private static void show_preferences() {
        CheeseMapConfig c = new CheeseMapConfig( toplevel, props );
        c.setVisible( true );
    }

    private static void load_properties() {
        props = new Properties();
        try {
            FileInputStream in = new FileInputStream( "cheesemap.cfg" );
            props.load( in );
            //props.list( System.out );
            in.close();
        }catch( Exception e ) {
            e.printStackTrace();
        }
    }

    public static void apply_properties() {
        // apply properties
        try {
            int loglevel = Integer.parseInt( props.getProperty( "loglevel" ) );
            CheeseLab.log_level = loglevel;
        }catch( Exception e ) {
            CheeseLab.log_level = 0;
        }
    }

    public static void save_properties() {
        try {
            FileOutputStream out = new FileOutputStream( "cheesemap.cfg" );
            props.store( out, null );
            out.close();
        }catch( Exception e ) {
            e.printStackTrace();
        }
    }

    /**
     * Set connection status of components that provide visual feed-back to user
     * @param connected  true if mapper is in connected state, false othervise.
     */
    private static void set_connect_status(boolean connected) {
        if (connected) {
            connect_action.putValue( SelectedAction.SHORT_DESCRIPTION, "Disconnect" );
            connect_action.putValue( SelectedAction.ACTION_COMMAND_KEY, "disconnect" );
            connect_action.putValue( SelectedAction.MNEMONIC_KEY,
                                     new Integer( KeyEvent.VK_D ) );
            connect_action.setSelected( true );
        }
        else {
            connect_action.putValue( SelectedAction.SHORT_DESCRIPTION, "Connect" );
            connect_action.putValue( SelectedAction.ACTION_COMMAND_KEY, "connect" );
            connect_action.putValue( SelectedAction.MNEMONIC_KEY,
                                     new Integer( KeyEvent.VK_C ) );
            connect_action.setSelected( false );
        }
    }

    private static void toggleToolbar( JToolBar bar, boolean visible ) {
        toggleToolbar( bar, visible, AKDockLayout.NORTH );
    }

    private static void toggleToolbar( JToolBar bar, boolean visible, Object c ) {
        Container content = toplevel.getContentPane();
        if ( visible ) {
            bar.setVisible( true );
            bar.setEnabled( true );
            // Check if constraint is ok, if not revert to NORTH
            Object [] constr_horiz = {
                                         AKDockLayout.NORTH, AKDockLayout.SOUTH,
                                     };
            Object [] constr_vert = {
                                        AKDockLayout.EAST, AKDockLayout.WEST
                                    };

            if( Arrays.asList( constr_horiz ).contains( c ) ) {
                bar.setOrientation( JToolBar.HORIZONTAL );
                content.add( bar, c );
            }
            else if( Arrays.asList( constr_vert ).contains( c ) ) {
                bar.setOrientation( JToolBar.VERTICAL );
                content.add( bar, c );
            }else{
                content.add( bar, AKDockLayout.NORTH );
            }
            content.validate();
        } else {
            // FIXME: If toolbar is undocked this beahves rather strangely
            Container parent = bar.getParent();
            bar.setVisible( false );
            bar.setEnabled( false );
            if( parent != null ) {
                parent.remove( bar );
                parent.validate();
            }
        }
    }

    private static void disconnect() {
        CheeseLab.dmesg( 2, "disconnect" );

        set_connect_status( false );

        if( sock != null ) {

            if( bozo != null ) bozo.running = false;

        try { in.close();	}catch( Exception e ) {	}
            try { out.close();	}catch( Exception e ) {	}
            try { sock.close();	}catch( Exception e ) {	}

            in = null;
            out = null;
            sock = null;
            bozo = null;
        }

        /* Don't send events if online status didn't change */
        if( !online )
            return;

        online = false;

        for( int i = 0; i < listeners.size(); i++ ) {
            ((CheeseMapListener)listeners.elementAt( i )).onlineChanged( online );
        }
    }

    private static void connect() {
        disconnect();
        connect_action.setSelected( true );
        connect_action.setSelected( false );

        // Make sure username is right
        apply_properties();

        String userid = "";
        String password = "";
        did_refresh = false;
        
        /* We load them from properties each time over and over again,
         * because the user might have canged them in the meantime.
         */
        userid = props.getProperty( "userid", "default" );
        password = props.getProperty( "password", "" );

        // Prompt for password if it's empty and userid is not default
        if( ( !userid.equals("default") && password.equals("") )
                || password.equals( "PROMPT" ) ) {
            // FIXME: This prompt shows the password to the screen :(
            password = JOptionPane.showInputDialog( toplevel, "Password" );
            if( password == null || password.equals("") ) {
                return;
            }
        }

        try {
            sock = new Socket( props.getProperty( "server" ), Integer.parseInt( props.getProperty( "port" ) ) );
            sock.setKeepAlive( true );

            in = new BufferedInputStream( sock.getInputStream() );
            out = new PrintStream( sock.getOutputStream(), true );
            bozo = new NetworkBonanza();

            CheeseLab.dmesg( 1, "Authenticating" );
            send_tx( 1, "auth", userid + ":" + password );

            set_connect_status ( true );

            cachedir = "cache/" +  props.getProperty( "server" ) + "_" + props.getProperty( "port" );
            // Shouldn't be needed anymore because map gets this event via other means
            //map.initcache();

            // Send our player tile
            canvas.set_follow_userid( userid );
            send_player_tile();
            canvas.set_follow_player( true );

        }catch( Exception e ) {
            CheeseLab.dmesg( 0, "Unable to connect: " + e );
            JOptionPane.showMessageDialog( toplevel, "Unable to connect: " + e );
            set_connect_status ( false );
        }
    }

    public static void send_player_tile() {
        CheeseFile f = new CheeseFile( "tiles/player_" + canvas.get_follow_userid() + ".png" );

        // Don't overwrite server tiles, so if you've set your avatar once it
        // it will work forever :p
        if( f.data != null )
            send_tx( 1, "tile", CheeseDecoder.encode_object( f ) );
    }

    public static void do_refresh() {
        // FIXME: Shouldn't be callable if not online
        if( JOptionPane.showConfirmDialog( toplevel, "Are you sure you want to sync? If the database is large and you have not sync'd before it may take a loooong time!", "Confirm Sync", JOptionPane.YES_NO_OPTION ) == JOptionPane.YES_OPTION ) {
            CheeseLab.dmesg( 2, "Performing pull of all new stuff: " + map.get_highest_mtime() );
            // Refresh should be mtime of last newest item in our cache
            doRefresh( true );
        }
    }

    public static int send_tx( int channel, String msgtype, String data ) {
        return( send_tx( channel, msgtype, data.getBytes() ) );
    }

    synchronized public static int send_tx( int channel, String msgtype, byte[] data ) {
        if( sock == null )
            return( -1 );

        last_sendid = CheeseTx.send_tx( out, last_sendid, channel, msgtype, data );
        return( 0 );
    }

    public static int handle_tx() {
        if( sock == null )
            return( -1 );

        CheeseLab.dmesg( 2, "Handle tx" );
        CheeseTx tx = CheeseTx.accept_tx( in );

        try {
            // Just pull the stuff from the tx as we did it manually before
            // TODO: make this nice
            if( tx == null ) {
                CheeseLab.dmesg( 2, "Got null tx, bye" );
                disconnect();
                return( -1 );
            }

            String msgtype = tx.msgtype;

            // FIXME: Those seem to be never used anyway
            //int seq = tx.sequence;
            //int channel = tx.channel;
            //String enc = tx.enc;
            //int len = tx.length;

            byte[] buffer = tx.buffer;

            CheeseLab.dmesg( 2, "msgtype is: " + msgtype );

            // Now based on msttype, pass the data to the proper routine
            if( msgtype.equals( "player" ) ) {
                CheesePlayerData p = (CheesePlayerData)CheeseDecoder.decode_object( buffer );
                CheesePlayerData op =  map.get_player( p.name );

                for( int i = 0; i < listeners.size(); i++ ) {
                    ((CheeseMapListener)listeners.elementAt( i )).playerChanged( p, op );
                }
            }else if( msgtype.equals( "progress" ) ) {
                // Display a progress message, if blank then no message to display
                String message = new String( tx.buffer );
                //CheeseLab.dmesg( 0, message );

                if( message.equals( "CLOSE" ) ) {
                    /* This is unbelieavable, but Swing seems to
                     * be unable to close the dialog if it gets show/hide events
                     * too fast. See SwingBug.java for demo. Geez!!!!!!!!!!!
                     */
                    Thread.sleep( 50 );
                    progress.setVisible( false );
                }else{
                    progress.set_message( message );
                    if( ! progress.isShowing() )
                        progress.setVisible( true );
                }
            }else if( msgtype.equals( "search" ) ) {
                // Search results
                find.set_search_results( buffer );
            }else if( msgtype.equals( "auth" ) ) {
                // Now we are connected and can do stuff
                online = true;
                for( int i = 0; i < listeners.size(); i++ ) {
                    ((CheeseMapListener)listeners.elementAt( i )).onlineChanged( online );
                }

                // Fire refresh
                if( CheeseMap.prop_get_boolean( "sync_on_world", true ) ) {
                    doRefresh( true );
                }
            }else if( msgtype.equals( "ping" ) ) {
                // pings allow network sync, set pingmonitor(message)=0
                // then wait for pingmonitor(message)=1, since
                // network ops happen in sequence your response will be complete
                // when you receieve your ping
                //setping( new String( buffer ) );
                setping();
            }else if( msgtype.equals( "tile" ) ) {
                // Accept tile, the format is tilename,tiledata
                CheeseFile f = (CheeseFile)CheeseDecoder.decode_object( tx.buffer );
                f.save( cachedir + "/tiles/" );
                CheeseLab.dmesg( 2, "Accepting tile: " + f.filename );

                for( int i = 0; i < listeners.size(); i++ ) {
                    ((CheeseMapListener)listeners.elementAt( i )).tileChanged( f.filename );
                }
            }else if( msgtype.equals( "message" ) ) {
                // Display message to user
                // TODO: need a chat thing too? :p
                CheeseLab.dmesg( 2, "Got message: " + new String( tx.buffer ) );
                JOptionPane.showMessageDialog( toplevel, new String( tx.buffer ) );
            }else if( msgtype.equals( "refresh" ) ) {
                // refresh complete
                did_refresh = true; // at least 1 refresh complete

                for( int i = 0; i < listeners.size(); i++ ) {
                    ((CheeseMapListener)listeners.elementAt( i )).refreshFinished();
                }
                progress.setVisible( false );
            }else if( msgtype.equals( "map" ) ) {
                // TODO: get this stuff to work with new event logic.
                map.update_map( buffer );
                canvas.map_changed();
                canvas.changed();
            }else if( msgtype.equals( "world" ) ) {
                // If using sync on world logic then only request room updates when they are available
                if( CheeseMap.prop_get_boolean( "sync_on_world", true ) )
                    // this will trigger map view refreshes
                    if( did_refresh ) { // did sync at least once using do_refresh
                        doRefresh( false );
                    }else{
                        // TODO: use this in drawing optimization
                        // TODO: revise in light of the new event handling system
                        canvas.world_changed();
                        canvas.remap();
                        canvas.map_changed();
                        canvas.changed();
                    }
            }else if( msgtype.equals( "room" ) ) {
                CheeseRoomData room = (CheeseRoomData)CheeseDecoder.decode_object( buffer );
                CheeseLab.dmesg( 2, "Decoded room: " + room.roomid + " - " +
                                 room.x + "," + room.y + "," + room.z + " - " + room.mtime );
                CheeseRoomData oldroom = map.get_room( room.roomid, 1, false );

                for( int i = 0; i < listeners.size(); i++ ) {
                    ((CheeseMapListener)listeners.elementAt( i )).roomChanged( room, oldroom );
                }
            }else{
                CheeseLab.dmesg( 2, "Unknown message type: " + msgtype );
            }
        }catch( Exception e ) {
            e.printStackTrace();
            disconnect();
        }
        return( 0 );
    }

    public static void setping() {
        pinger = true;
    }

    // WARNING: due to syncro issues there is only 1 pinger now
    public static boolean checkping() {
        if( sock == null )
            return( true );

        return( pinger );
    }

    public static void clearping( String msg ) {
        pinger = false;
    }

    public static void set_edit_tool( String toolname ) {
  /*      if (toolname.equals("move_tool")) {
            JOptionPane.showMessageDialog( toplevel, "Not done, sorry!\nUse keyboard to move rooms around for now." );
            toolbox.setActiveTool( "select_tool" );
        }
        else*/
            canvas.setEditTool (toolname);
    }

    public static void main(String[] args) {
        //Schedule a job for the event-dispatching thread:
        //creating and showing this application's GUI.

        // Set LAF based on OS
        String os = System.getProperty( "os.name" );
        String LAF = null;
        System.out.println( "Finding LAF for " + os );
        if( os.toLowerCase().matches( "windows.*" ) ) {
            System.out.println( "Setting windows LAF" );
            LAF = "com.sun.java.swing.plaf.windows.WindowsLookAndFeel";
        /* GTK LAF looks bad currently with cheeselab
        }else if( os.toLowerCase().matches( "linux" ) ) {
            System.out.println( "GTK LAF" );
            LAF = "com.sun.java.swing.plaf.gtk.GTKLookAndFeel";
        */
        }else{
            System.out.println( "Default LAF" );
            LAF = UIManager.getCrossPlatformLookAndFeelClassName();
        }

        if( LAF != null ) { 
            try {
                System.out.println( "Setting LAF to " + LAF );
                UIManager.setLookAndFeel( LAF );
            } catch (Exception e) { }
        }

        // Make sure that windows behave "natively" from Java 1.5 onwards
        // System.setProperty( "java.awt.Window.locationByPlatform", "true" );

        javax.swing.SwingUtilities.invokeLater( new Runnable() {
                                                    public void run() {
                                                        createAndShowGUI();
                                                        CheeseLab.dmesg( 2, "End of main_run()" );
                                                    }
                                                } );
    }

    public static int get_tile_size() {
        Dimension d = canvas.get_tile_size();
        return (d.width < d.height ) ? d.height : d.width;
    }

    public static void set_tile_size( int size ) {
        canvas.set_tile_size( size, size );
    }

    public static void change_avatar() {
        JFileChooser chooser = new JFileChooser();

        chooser.showDialog( toplevel, "Select Image" );
        File f = chooser.getSelectedFile();

        if( f != null ) {
            // Read the image, and save it out.  If we are online, also push it to the server
            CheeseLab.dmesg( 0, "They chose: " + f );
            ImageIcon ii = new ImageIcon( f.getAbsolutePath() );

            if( ii.getIconWidth() == -1 ) {
                // Error
                JOptionPane.showMessageDialog( toplevel, "Error loading image" );
            }else{
                Image i = ii.getImage().getScaledInstance( 128, 128, Image.SCALE_SMOOTH );

                // Force loading
                while( ! canvas.prepareImage( i, -1, -1, (ImageObserver)canvas ) );

                // Resize
                BufferedImage im = new BufferedImage( 128, 128, Transparency.TRANSLUCENT );
                Graphics2D g = im.createGraphics();
                g.drawImage( i, 0, 0, null );
                g.dispose();

                Iterator writers = ImageIO.getImageWritersByFormatName( "png" );
                ImageWriter writer = (ImageWriter)writers.next();

                File output = new File( "tiles/player_" + canvas.get_follow_userid() + ".png" );
                try {
                    ImageOutputStream ios = ImageIO.createImageOutputStream( output );
                    writer.setOutput( ios );
                    writer.write( im );

                    ios.close();

                    // Send tile change event
                    for( int j = 0; j < listeners.size(); j++ ) {
                        ((CheeseMapListener)listeners.elementAt( j )).
                        tileChanged( "player_" + canvas.get_follow_userid() + ".png" );
                    }

                    if( sock != null )
                        send_player_tile();

                    JOptionPane.showMessageDialog( toplevel, "Icon changed" );
                }catch( Exception e ) {
                    e.printStackTrace();
                }
            }
        }
    }

    /**
     * Exit cleanly
     */
    public static void exit() {
        disconnect();
        save_ui_prefs();
        System.exit( 0 );
    }

    /**
     * Save UI preferences: window size and position,
     * tile size etc.
     */
    public static void save_ui_prefs() {
        Rectangle tlsiz = toplevel.getBounds();
        props.setProperty( "mainwin_x", Integer.toString( tlsiz.x ) );
        props.setProperty( "mainwin_y", Integer.toString( tlsiz.y ) );
        props.setProperty( "mainwin_w", Integer.toString( tlsiz.width ) );
        props.setProperty( "mainwin_h", Integer.toString( tlsiz.height ) );
        props.setProperty( "mainwin_ontop", ontop_action.isSelected() ? "true" : "false" );
        props.setProperty( "tile_size", Integer.toString( get_tile_size() ) );
        props.setProperty( "render_grid", canvas.get_render_grid() ? "true" : "false" );
        props.setProperty( "render_notes", canvas.get_render_notes() ? "true" : "false" );
        props.setProperty( "render_futures", canvas.get_render_future() ? "true" : "false" );
        props.setProperty( "show_navigator", navi_action.isSelected() ? "true" : "false" );

        // Save toolbars state and position
        props.setProperty( "show_toolbar", toolbar.isVisible() ? "true" : "false" );
        props.setProperty( "show_toolbox", toolbox.isVisible() ? "true" : "false" );
        // TODO: Implement floating toolbar state and position saving
        AKDockLayout topcont = (AKDockLayout)toplevel.getContentPane().getLayout();
        String c = (String)topcont.getConstraints( toolbar );
        if( c != null )
            props.setProperty( "toolbar_location", c );
        c = (String)topcont.getConstraints( toolbox );
        if( c != null )
            props.setProperty( "toolbox_location", c );

        navi.save_ui_prefs();

        // TODO: Save/Restore navigator window position too
        CheeseMap.save_properties();
    }

    /* Global property getting routines that perform some cleaning
     * and normal data conversion, e.g. boolean can be 1,0,true,false,yes,no,etc */
    public static boolean prop_get_boolean( String name ) {
        return prop_get_boolean( name, false );
    }

    public static boolean prop_get_boolean( String name, boolean default_value ) {
        String value = props.getProperty( name );

        if( value == null )
            return( default_value );

        value = value.toLowerCase();
        if( value.equals( "1" ) || value.equals( "true" ) || value.equals( "yes" ) )
            return( true );

        // Anything else is false
        return( false );

    }

    public static String prop_get_string( String name ) {
        return prop_get_string( name, "" );
    }

    public static String prop_get_string( String name, String default_value ) {
        String value = props.getProperty( name );

        if( value == null )
            value = default_value;

        return( value );
    }

    public static int prop_get_int( String name ) {
        return prop_get_int( name, 0 );
    }

    public static int prop_get_int( String name, int default_value ) {
        String value = props.getProperty( name );
        int retval = default_value;

        if( value == null )
            return( retval );

        try {
            retval = Integer.parseInt( value );
        }catch( NumberFormatException e ) {}

        return( retval );
    }

    public static void set_follow_player( boolean f ) {
        canvas.set_follow_player( f );
    }

    public static void set_follow_userid( String userid ) {
        canvas.set_follow_userid( userid );
    }

    public static boolean get_follow_player() {
        return canvas.get_follow_player();
    }

    public static String get_follow_userid() {
        return canvas.get_follow_userid();
    }

    public static void set_origin( int x , int y , int z ) {
        canvas.set_origin( x, y, z );
    }

    public static void addCheeseMapListener( CheeseMapListener listener ) {
        listeners.add( listener );
    }

    public static void removeCheeseMapListener( CheeseMapListener listener ) {
        listeners.remove( listener );
    }

    public static void doRefresh(boolean progress) {
        // Tell listeners not to redraw rapidly
        for( int i = 0; i < listeners.size(); i++ ) {
            ((CheeseMapListener)listeners.elementAt( i )).refreshStarted();
        }

        // Send refresh request to server
        send_tx( 1, "refresh", Long.toString( map.get_highest_mtime() ) +
                 ( progress ? "" : " noprogress" ) );
    }
}
